'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var borel = require('borel');
var veho = require('veho');
var farbe = require('farbe');
var hatsu = require('hatsu');
var typen = require('typen');
var palett = require('palett');

/**
 * Create a chalk from a hsl array
 * @param {[number,number,number]} hsl
 * @returns {function}
 */

const tube = hsl => hatsu.Hatsu.hsl(hsl);

const {
  isNumeric: isNum
} = typen.Num;

const amp = (x, m, rto, min) => (x - m) * rto + min;
const makeHatsu = (x, m, [rH, rS, rL], [mH, mS, mL]) => {
  var _ref;

  return _ref = [amp(x, m, rH, mH), amp(x, m, rS, mS), amp(x, m, rL, mL)], tube(_ref);
};

const {
  mutateMap,
  map: mapAr
} = veho.Ar;

/**
 *
 * @param {*[]} ar
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const vecVisual = (ar, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap : mapAr;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min;

    const soleChalk = (_min = min, tube(_min));
    return retFn ? map(ar, x => isNum(x) ? soleChalk : naChalk) : map(ar, x => {
      var _x, _x2;

      return isNum(x) ? (_x = x, soleChalk(_x)) : (_x2 = x, naChalk(_x2));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? map(ar, x => isNum(x) ? makeHatsu(x, m, rto, min) : naChalk) : map(ar, x => {
    var _x3, _x4;

    return isNum(x) ? (_x3 = x, makeHatsu(x, m, rto, min)(_x3)) : (_x4 = x, naChalk(_x4));
  });
};

/**
 *
 * @param {*[][]} mx
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const mxVisual = (mx, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap : mapAr;

  if (isNaN(d) || !d || dif.every(el => !el)) {
    var _min;

    const soleChalk = (_min = min, tube(_min));
    return retFn ? map(mx, el => isNum(el) ? soleChalk : naChalk) : map(mx, el => {
      var _el, _el2;

      return isNum(el) ? (_el = el, soleChalk(_el)) : (_el2 = el, naChalk(_el2));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? map(mx, el => isNum(el) ? makeHatsu(el, m, rto, min) : naChalk) : map(mx, el => {
    var _el3, _el4;

    return isNum(el) ? (_el3 = el, makeHatsu(el, m, rto, min)(_el3)) : (_el4 = el, naChalk(_el4));
  });
};
/**
 *
 * @param {*[][]} mx
 * @param {number} y
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const mxVisualCol = (mx, y, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const mapCol = mutate ? veho.Mx.mutateCol : veho.Mx.mapCol;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min2;

    const soleChalk = (_min2 = min, tube(_min2));
    return retFn ? mapCol(mx, y, el => isNum(el) ? soleChalk : naChalk) : mapCol(mx, y, el => {
      var _el5, _el6;

      return isNum(el) ? (_el5 = el, soleChalk(_el5)) : (_el6 = el, naChalk(_el6));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? mapCol(mx, y, el => isNum(el) ? makeHatsu(el, m, rto, min) : naChalk) : mapCol(mx, y, el => {
    var _el7, _el8;

    return isNum(el) ? (_el7 = el, makeHatsu(el, m, rto, min)(_el7)) : (_el8 = el, naChalk(_el8));
  });
};

const marks = {
  fresh: {
    max: palett.Palett.lightGreen.accent_3,
    min: palett.Palett.red.accent_2,
    na: palett.Greys.brown.base
  },
  subtle: {
    max: palett.Greys.grey.lighten_5,
    min: palett.Greys.grey.darken_1,
    na: palett.Palett.indigo.lighten_2
  }
};

const {
  fromHex,
  toHsl
} = farbe.Rgb;

const parseHsl = color => {
  var _ref, _color;

  return typeof color === 'string' ? (_ref = (_color = color, fromHex(_color)), toHsl(_ref)) : color;
};
/**
 *
 * @param max
 * @param min
 * @returns {{dif: [number,number,number], min: [number,number,number]}}
 */


const colorBound = (max, min) => ({
  min,
  dif: veho.Ar.mutateMap(max, (a, i) => a - min[i], 3)
});

class Visual {
  /**
   *
   * @param {*[]} arr
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   */
  static vector(arr, {
    mark: {
      max = marks.fresh.max,
      min = marks.fresh.min,
      na = marks.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max, _min, _ref2, _na;

    return vecVisual(arr, {
      val: borel.Stat.bound(arr, {
        dif: true,
        level: 2
      }),
      pal: colorBound((_max = max, parseHsl(_max)), (_min = min, parseHsl(_min))),
      naChalk: (_ref2 = (_na = na, parseHsl(_na)), tube(_ref2)),
      mutate,
      retFn
    });
  }
  /**
   *
   * @param {*[][]} mx
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {number} [direct=1] - 0:p-wise, 1:r-wise, 2:c-wise
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   * @returns {*}
   */


  static matrix(mx, {
    mark: {
      max = marks.subtle.max,
      min = marks.subtle.min,
      na = marks.subtle.na
    } = {},
    direct = 1,
    mutate = false,
    retFn = false
  } = {}) {
    var _max2, _min2, _ref3, _na2;

    if (!veho.Mx.isMat(mx)) throw new Error('Not valid matrix');

    switch (direct) {
      case 2:
        const mapCols = veho.Mx.mapColumns;
        return mapCols(mx, c => Visual.vector(c, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 1:
        const mapRows = mutate ? mutateMap : mapAr;
        return mapRows(mx, r => Visual.vector(r, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 0:
      default:
        return mxVisual(mx, {
          val: borel.StatMx.bound(mx, {
            dif: true,
            level: 2
          }),
          pal: colorBound((_max2 = max, parseHsl(_max2)), (_min2 = min, parseHsl(_min2))),
          naChalk: (_ref3 = (_na2 = na, parseHsl(_na2)), tube(_ref3)),
          mutate,
          retFn
        });
    }
  }
  /**
   *
   * @param {*[][]} mx
   * @param {number} y
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} deep
   */


  static column(mx, y, {
    mark: {
      max = marks.fresh.max,
      min = marks.fresh.min,
      na = marks.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max3, _min3, _ref4, _na3;

    return mxVisualCol(mx, y, {
      val: borel.StatMx.boundCol(mx, y, {
        dif: true,
        level: 2
      }),
      pal: colorBound((_max3 = max, parseHsl(_max3)), (_min3 = min, parseHsl(_min3))),
      naChalk: (_ref4 = (_na3 = na, parseHsl(_na3)), tube(_ref4)),
      mutate,
      retFn
    });
  }

}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

var _Object$entries$map;
class Pal {}

_defineProperty(Pal, "bases", (_Object$entries$map = Object.entries(palett.Palett).map(([name, {
  base
}]) => [name, base]), veho.Ob.fromEntries(_Object$entries$map)));

_defineProperty(Pal, "marks", marks);

/**
 *
 * @param {*} x
 * @return {string}
 */


const totx = x => `${x}`;

const isTab = c => c === '\t' || c === ' ';

const tabify = tx => {
  var _tx;

  const i = (_tx = tx, deNaTab(_tx));
  return endsBracs(tx) ? tx : `${tx.substring(0, i)}[${tx.substring(i)}]`;
};

const deNaTab = tx => {
  let i = 0;

  for (let {
    length
  } = tx; i < length; i++) if (!isTab(tx.charAt(i))) return i;

  return i;
};

const endsBracs = tx => tx.endsWith(')') || tx.endsWith(']');

const Presets = {
  fresh: {
    max: palett.Palett.lightGreen.accent_3,
    min: palett.Palett.red.accent_2,
    na: palett.Greys.brown.base
  },
  subtle: {
    max: palett.Greys.grey.lighten_5,
    min: palett.Greys.grey.darken_1,
    na: palett.Palett.indigo.lighten_2
  }
};

const FRESH = {
  max: palett.Palett.lightGreen.accent_3,
  min: palett.Palett.red.accent_2,
  na: palett.Palett.brown.base
};
const SUBTLE = {
  max: palett.Palett.grey.lighten_5,
  min: palett.Palett.grey.darken_1,
  na: palett.Palett.indigo.lighten_2
};
const GLACIAL = {
  max: palett.Palett.teal.accent_2,
  min: palett.Palett.blue.darken_3,
  na: palett.Palett.brown.accent_3
};
const METRO = {
  max: palett.Palett.pink.darken_1,
  min: palett.Palett.blueGrey.accent_3,
  na: palett.Palett.teal.darken_4
};

const Presets$1 = {
  fresh: {
    max: palett.Palett.lightGreen.accent_3,
    min: palett.Palett.red.accent_2,
    na: palett.Greys.brown.base
  },
  subtle: {
    max: palett.Greys.grey.lighten_5,
    min: palett.Greys.grey.darken_1,
    na: palett.Palett.indigo.lighten_2
  }
};

const {
  hsl
} = hatsu.Hatsu;
const L = '{ ',
      R = ' }';
const Tubes = {
  0: hsl([45, 100, 53]),
  1: hsl([44, 100, 59]),
  2: hsl([43, 100, 64]),
  3: hsl([42, 100, 70]),
  4: hsl([41, 100, 74]),
  5: hsl([40, 100, 78]),
  6: hsl([39, 100, 82])
};
const Puncs = veho.Ob.mapValues(Tubes, hsl => {
  var _L, _R;

  return [(_L = L, hsl(_L)), (_R = R, hsl(_R))];
});
const BRC = veho.Ob.mapValues(Puncs, ([L, R]) => content => L + content + R);
const {
  hsl: hsl$1
} = hatsu.Hatsu;
const L$1 = '[ ',
      R$1 = ' ]';
const Tubes$1 = {
  0: hsl$1([199, 100, 63]),
  1: hsl$1([201, 100, 68]),
  2: hsl$1([203, 100, 72]),
  3: hsl$1([205, 100, 76]),
  5: hsl$1([207, 100, 84]),
  4: hsl$1([209, 100, 80]),
  6: hsl$1([211, 100, 88])
};
const Puncs$1 = veho.Ob.mapValues(Tubes$1, hsl => {
  var _L, _R;

  return [(_L = L$1, hsl(_L)), (_R = R$1, hsl(_R))];
});
const BRK = veho.Ob.mapValues(Puncs$1, ([L, R]) => content => L + content + R);
const {
  hex
} = hatsu.Hatsu;
/**
 *
 * @type {Object<string,function>}
 */

const PAL = {
  IDX: hex(palett.Greys.brown.lighten_5),
  STR: hex(palett.Palett.lightGreen.accent_2),
  NUM: hex(palett.Palett.deepOrange.accent_2),
  BOO: hex(palett.Palett.teal.lighten_2),
  UDF: hex(palett.Palett.brown.lighten_3),
  BRK: hex(palett.Palett.blue.accent_2),
  BRC: hex(palett.Palett.amber.base),
  FNC: hex(palett.Palett.green.accent_4)
};
const {
  hsl: hsl$2
} = hatsu.Hatsu;
const IDX = {
  0: hsl$2([20, 16, 93]),
  1: hsl$2([18, 18, 88]),
  2: hsl$2([17, 20, 83]),
  3: hsl$2([16, 22, 78]),
  4: hsl$2([15, 24, 73]),
  5: hsl$2([14, 26, 69]),
  6: hsl$2([14, 28, 65])
};

function _defineProperty$1(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const {
  hex: hex$1
} = hatsu.Hatsu;
const Thm = {
  brm: hex$1(palett.Palett.blueGrey.base),
  br: hex$1(palett.Palett.orange.lighten_3),
  pr: hex$1(palett.Palett.indigo.lighten_1)
};

const bracketMain = tx => Thm.brm('[') + tx + Thm.brm(']');

const bracket = tx => Thm.br('[') + tx + Thm.br(']');

const parenthesis = tx => Thm.pr('(') + tx + Thm.pr(')');

const render = (tx, {
  indent,
  stream
}) => {
  if (tx === null || tx === void 0 ? void 0 : tx.length) stream.push(tx);
  return ' '.repeat(indent << 1) + stream.join(' ');
};

const preset = (label, ...items) => {
  var _label;

  let stream = [],
      indent,
      len;

  if (label && (label = String(label)) && (len = label.length) && (indent = (_label = label, deNaTab(_label))) < len) {
    var _label$slice;

    stream.push((_label$slice = label.slice(indent), bracketMain(_label$slice)));
  }

  if (items.length) {
    var _items$map$join;

    stream.push((_items$map$join = items.map(totx).join(','), parenthesis(_items$map$join)));
  }

  return {
    indent,
    stream
  };
};

let _Symbol$toPrimitive;

class Callable extends Function {
  constructor(f) {
    super();
    Reflect.setPrototypeOf(f, new.target.prototype);
    return f;
  }

}
/**
 * @type {object<string,string>}
 */


_Symbol$toPrimitive = Symbol.toPrimitive;

class Ink extends Callable {
  /** @type {number} */

  /** @type {string[]} */
  constructor(label, ...items) {
    super(tx => render(tx, this));

    _defineProperty$1(this, "indent", void 0);

    _defineProperty$1(this, "stream", void 0);

    this.cr(label, ...items);
    return new Proxy(this, {
      /** @returns {Ink|function(...string):Ink} */
      get(target, p, receiver) {
        var _String;

        if (p in target) return target[p];
        target.stream.push((_String = String(p), bracket(_String)));
        return (...items) => {
          var _items$map$join;

          target.stream.push((_items$map$join = items.map(totx).join(', '), parenthesis(_items$map$join)));
          return receiver;
        };
      }

    });
  }

  [_Symbol$toPrimitive](h) {
    switch (h) {
      case 'string':
      case 'default':
        return render(null, this);

      case 'number':
        return this.indent;

      default:
        throw new Error('ink Symbol.toPrimitive error');
    }
  }

  cr(label, ...items) {
    const {
      indent,
      stream
    } = preset(label, ...items);
    this.indent = indent;
    this.stream = stream;
    return this;
  }

  asc() {
    this.indent++;
    return this;
  }

  desc() {
    this.indent--;
    return this;
  }

  tag(key, ...items) {
    var _ref, _key, _items$map$join2;

    this.stream.push((_ref = (_key = key, totx(_key)), tabify(_ref)));
    if (items.length) this.stream.push((_items$map$join2 = items.map(totx).join(','), parenthesis(_items$map$join2)));
    return this;
  }

  br(...items) {
    this.stream.push(items.map(parenthesis).join(','));
    return this;
  }

  p(...items) {
    this.stream.push(...items);
    return this;
  }

  get tx() {
    return render(null, this);
  }

  get say() {
    return render(null, this);
  }

  toString() {
    return render(null, this);
  }

}

const ink = new Ink();

function _defineProperty$2(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}
/**
 * If y >= 0 then sort rows, else (e.g. y===undefined) sort keys.
 * @param {*[]} keys
 * @param {*[][]} rows
 * @param {function(*,*):number} comparer
 * @param {number} [y]
 * @returns {[*[], *[][]]}
 */


const sortAlong = (keys, rows, comparer, y) => {
  const comp = (a, b) => comparer(a[0], b[0]),
        ks = Array(keys.length),
        rs = y >= 0 ? rows.map((r, i) => [r[y], keys[i], r]).sort(comp).map(([, k, r], i) => {
    ks[i] = k;
    return r;
  }) : rows.map((r, i) => [keys[i], r]).sort(comp).map(([k, r], i) => {
    ks[i] = k;
    return r;
  });

  return [ks, rs];
};

class CrosTab {
  /**
   *
   * @param {*[]} side
   * @param {*[]} banner
   * @param {*[][]} matrix
   * @param {string} [title]
   */
  constructor(side, banner, matrix, title) {
    _defineProperty$2(this, "side", void 0);

    _defineProperty$2(this, "banner", void 0);

    _defineProperty$2(this, "matrix", void 0);

    _defineProperty$2(this, "title", void 0);

    this.side = side;
    this.banner = banner;
    this.matrix = matrix;
    this.title = title || '';
  }
  /**
   * Shallow copy
   * @param {*[]} side
   * @param {*[]} banner
   * @param {*[][]} matrix
   * @param {string} [title]
   * @return {CrosTab}
   */


  static from({
    side,
    banner,
    matrix,
    title
  }) {
    return new CrosTab(side, banner, matrix, title);
  }
  /**
   * Shallow copy
   * @param {*[]} side
   * @param {*[]} banner
   * @param {function(number,number):*} func
   * @param {string} [title]
   * @return {CrosTab}
   */


  static ini({
    side,
    banner,
    func,
    title
  }) {
    const matrix = veho.Mx.ini(side === null || side === void 0 ? void 0 : side.length, banner === null || banner === void 0 ? void 0 : banner.length, (x, y) => func(x, y));
    return CrosTab.from({
      side,
      banner,
      matrix,
      title
    });
  }
  /**
   *
   * @param {*[]|boolean} [side]
   * @param {*[]|boolean} [banner]
   * @returns {*}
   */


  toSamples({
    side,
    banner
  }) {
    const [sideIsArr, bannerIsArr] = [Array.isArray(side) && side.length || side === true, Array.isArray(banner) && banner.length || banner === true];

    if (sideIsArr) {
      if (bannerIsArr) {
        var _side$map;

        const entries = banner.map(x => [x, this.coin(x)]);
        return (_side$map = side.map(x => this.matrix[this.roin(x)]), veho.Mx.transpose(_side$map)).map(col => veho.Ob.fromEntries(entries, i => col[i]));
      } else {
        var _this$matrix;

        const entries = side.map(x => [x, this.roin(x)]);
        return (_this$matrix = this.matrix, veho.Mx.transpose(_this$matrix)).map(col => veho.Ob.fromEntries(entries, i => col[i]));
      }
    } else {
      if (bannerIsArr) {
        const entries = banner.map(x => [x, this.coin(x)]);
        return this.matrix.map(row => veho.Ob.fromEntries(entries, i => row[i]));
      } else {
        return veho.Samples.fromCrosTab(this);
      }
    }
  }

  get toJson() {
    var _this$matrix2;

    return {
      side: this.side.slice(),
      banner: this.banner.slice(),
      matrix: (_this$matrix2 = this.matrix, veho.clone(_this$matrix2)),
      title: this.title
    };
  }

  get columns() {
    return veho.Mx.transpose(this.matrix);
  }
  /**
   *
   * @param {*[][]} [newMatrix]
   * @param {*[]} [newSide]
   * @param {*[]} [newBanner]
   * @returns {CrosTab}
   */


  reboot(newMatrix, newSide, newBanner) {
    if (newMatrix) this.matrix = newMatrix;
    if (newSide) this.side = newSide;
    if (newBanner) this.banner = newBanner;
    return this;
  }
  /**
   * Shallow copy
   * @param {*[][]} [newMatrix]
   * @param {*[]} [newSide]
   * @param {*[]} [newBanner]
   * @return {CrosTab}
   */


  clone(newMatrix, newSide, newBanner) {
    return new CrosTab(newSide || this.side.slice(), newBanner || this.banner.slice(), newMatrix || veho.Mx.clone(this.matrix), this.title);
  }

  get size() {
    return veho.Mx.size(this.matrix);
  }

  get ht() {
    var _this$side;

    return (_this$side = this.side) === null || _this$side === void 0 ? void 0 : _this$side.length;
  }

  get wd() {
    var _this$banner;

    return (_this$banner = this.banner) === null || _this$banner === void 0 ? void 0 : _this$banner.length;
  }

  cell(x, y) {
    return this.matrix[x][y];
  }

  queryCell(r, c) {
    const x = this.side.indexOf(r),
          y = this.banner.indexOf(c);
    return x < 0 || y < 0 ? null : this.matrix[x][y];
  }

  queryCoordinate(r, c) {
    return {
      x: this.side.indexOf(r),
      y: this.banner.indexOf(c)
    };
  }

  roin(r) {
    return Number.isInteger(r) && 0 <= r && r < this.ht ? r : this.side.indexOf(r);
  }

  coin(c) {
    return Number.isInteger(c) && 0 <= c && c < this.wd ? c : this.banner.indexOf(c);
  }

  row(r) {
    const x = this.roin(r);
    return this.matrix[x];
  }

  column(c) {
    const y = this.coin(c);
    return this.matrix.map(row => row[y]);
  }

  setRow(r, newRow) {
    const x = this.roin(r);
    this.matrix[x] = newRow;
    return this;
  }

  setColumn(c, newColumn) {
    const y = this.coin(c);
    this.matrix.forEach((row, i) => {
      row[y] = newColumn[i];
    });
    return this;
  }

  map(fn, {
    mutate = true
  } = {}) {
    const matrix = veho.Mx.map(this.matrix, fn);
    return mutate ? this.reboot(matrix) : this.clone(matrix);
  }
  /**
   * Push row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} row
   * @return {CrosTab}
   */


  pushRow(field, row) {
    this.side.push(field);
    this.matrix.push(row);
    return this;
  }
  /**
   * Unshift row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} row
   * @return {CrosTab}
   */


  unshiftRow(field, row) {
    this.side.unshift(field);
    this.matrix.unshift(row);
    return this;
  }
  /**
   * Push column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} col
   * @return {CrosTab}
   */


  pushCol(field, col) {
    this.banner.push(field);
    this.matrix.forEach((row, i) => row.push(col[i]));
    return this;
  }
  /**
   * Unshift column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} col
   * @return {CrosTab}
   */


  unshiftCol(field, col) {
    this.banner.unshift(field);
    this.matrix.forEach((row, i) => row.unshift(col[i]));
    return this;
  }

  select({
    side: s,
    banner: b
  } = {}, {
    mutate = true
  } = {}) {
    let {
      matrix,
      side,
      banner
    } = this;

    if (s && s.length) {
      const roins = s.map(x => this.roin(x));
      matrix = veho.Ar.select(matrix, roins);
      side = s;
    }

    if (b && b.length) {
      const coins = b.map(x => this.coin(x));
      matrix = veho.Mx.select(matrix, coins);
      banner = b;
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  slice({
    rows = {
      begin: undefined,
      end: undefined
    },
    cols = {
      begin: undefined,
      end: undefined
    }
  }, {
    mutate = true
  } = {}) {
    let {
      side,
      banner,
      matrix
    } = this;
    let begin, end;

    if (rows) {
      ({
        begin,
        end
      } = rows);
      side = side.slice(begin, end);
      matrix = matrix.slice(begin, end);
    }

    if (cols) {
      ({
        begin,
        end
      } = cols);
      banner = this.banner.slice(begin, end);
      matrix = matrix.map(row => row.slice(begin, end));
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  sort(by = 'rows', field, comparer, {
    mutate = true
  } = {}) {
    var _matrix, _matrix2;

    let {
      side,
      banner,
      matrix
    } = this;

    switch (by.charAt(0)) {
      case 'c':
        const x = this.roin(field);
        [banner, matrix] = sortAlong(banner, (_matrix = matrix, veho.Mx.transpose(_matrix)), comparer, x);
        matrix = (_matrix2 = matrix, veho.Mx.transpose(_matrix2));
        break;

      case 'r':
      default:
        const y = this.coin(field);
        [side, matrix] = sortAlong(side, matrix, comparer, y);
        break;
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  sortLabel(by = 'rows', comparer) {
    var _matrix3, _matrix4;

    let {
      side,
      banner,
      matrix,
      title
    } = this;

    switch (by.charAt(0)) {
      case 'c':
        side = side.slice();
        [banner, matrix] = sortAlong(banner, (_matrix3 = matrix, veho.Mx.transpose(_matrix3)), comparer);
        matrix = (_matrix4 = matrix, veho.Mx.transpose(_matrix4));
        break;

      case 'r':
      default:
        banner = banner.slice();
        [side, matrix] = sortAlong(side, matrix, comparer);
        break;
    }

    return new CrosTab(side, banner, matrix, title);
  }

  transpose(newTitle, {
    mutate = true
  } = {}) {
    const {
      banner: side,
      side: banner,
      columns: matrix
    } = this;
    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

}
const {
  spliceCols,
  size
} = veho.Mx;
const {
  select,
  ini: ar
} = veho.Ar;

const locInd = (mx, ys, vs, hi) => mx.findIndex(row => {
  for (let i = 0; i < hi; i++) if (row[ys[i]] !== vs[i]) return false;

  return true;
});

class MxJoin {}

_defineProperty$2(MxJoin, "joinUnion", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hL = mxL.length,
        hR = mxR.length,
        hi = ysR.length,
        mx = Array(hL),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        [, wL] = size(mxL2),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending)),
        [, wR] = size(mxR2),
        idxRiLs = new Set();

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);

    if (x < 0) {
      mx[i] = vsL.concat(mxL2[i], ar(wR, fillEmpty));
    } else {
      mx[i] = vsL.concat(mxL2[i], mxR2[x]);
      idxRiLs.add(x);
    }
  }

  for (let i = 0, x, vsR; i < hR; i++) {
    var _vsR$concat;

    if (idxRiLs.has(i)) {
      continue;
    }

    vsR = select(mxR[i], ysR, hi);
    x = locInd(mxL, ysL, vsR, hi);
    if (x < 0) _vsR$concat = vsR.concat(ar(wL, fillEmpty), mxR2[i]), mx.push(_vsR$concat);
  }

  return mx;
});

_defineProperty$2(MxJoin, "joinLeft", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hL = mxL.length,
        hi = ysL.length,
        mx = Array(hL),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending)),
        [, wR] = size(mxR2);

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);
    mx[i] = x < 0 ? vsL.concat(mxL2[i], ar(wR, fillEmpty)) : vsL.concat(mxL2[i], mxR2[x]);
  }

  return mx;
});

_defineProperty$2(MxJoin, "joinRight", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hR = mxR.length,
        hi = ysR.length,
        mx = Array(hR),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        [, wL] = size(mxL2),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending));

  for (let i = 0, x, vsR; i < hR; i++) {
    vsR = select(mxR[i], ysR, hi);
    x = locInd(mxL, ysL, vsR, hi);
    mx[i] = x < 0 ? vsR.concat(ar(wL, fillEmpty), mxR2[i]) : vsR.concat(mxL2[x], mxR2[i]);
  }

  return mx;
});

_defineProperty$2(MxJoin, "joinIntersect", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}) => {
  const hL = mxL.length,
        hi = ysL.length,
        mx = [],
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending));

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);
    if (x < 0) continue;
    mx.push(vsL.concat(mxL2[i], mxR2[x]));
  }

  return mx;
});

const pivotMode = stat => {
  return typeof stat === 'string' ? veho.PivotModes[stat] || (stat === 'cnt' ? veho.PivotModes.count : veho.PivotModes.array) : veho.PivotModes.array;
};

veho.Pivot.prototype.pivotStat = function ([x, y], [index, stat], {
  boot
} = {}) {
  var _stat;

  const mode = (_stat = stat, pivotMode(_stat));
  let {
    side,
    banner,
    matrix
  } = this.pivot([x, y, index], {
    mode,
    boot
  });
  if (mode === veho.PivotModes.array) matrix = veho.Mx.map(matrix, stat);
  return {
    side,
    banner,
    matrix
  };
};
/**
 *
 * @param {number} x
 * @param {number} y
 * @param {{index:number,stat:string|function(*[]):number}[]}fields
 */


veho.Pivot.prototype.pivotMatrices = function ([x, y], fields) {
  const hi = fields.length;
  let {
    side,
    banner,
    matrix
  } = this.pivotStat([x, y], fields[0], {
    boot: true
  });
  const matrices = [matrix];

  for (let i = 1; i < hi; i++) {
    var _this$pivotStat$matri;

    _this$pivotStat$matri = this.pivotStat([x, y], fields[i], {
      boot: false
    }).matrix, matrices.push(_this$pivotStat$matri);
  }

  return {
    side,
    banner,
    matrix: veho.Mx.zip(matrices)
  };
};

// const { random } = Math
const {
  random,
  abs,
  exp,
  log,
  sqrt,
  pow,
  cos,
  sin,
  PI
} = Math;

const randIdx = arr => ~~(random() * arr.length);

const randSel = arr => arr[randIdx(arr)];

const R0 = 3.442619855899;
const R1 = 1.0 / R0;
const R0S = exp(-0.5 * R0 * R0);
const N2P32 = -pow(2, 32);
const M1 = 2147483648.0;
const VN = 9.91256303526217e-3;

class Ziggurat {
  constructor(mean = 0, stdev = 1) {
    this.jsr = 123456789;
    this.wn = Array(128);
    this.fn = Array(128);
    this.kn = Array(128);
    this.mean = mean;
    this.stdev = stdev;
    this.preset();
  }

  preset() {
    // seed generator based on current time
    this.jsr ^= new Date().getTime();
    let m1 = M1,
        dn = R0,
        tn = R0,
        vn = VN,
        q = vn / R0S;
    this.kn[0] = ~~(dn / q * m1);
    this.kn[1] = 0;
    this.wn[0] = q / m1;
    this.wn[127] = dn / m1;
    this.fn[0] = 1.0;
    this.fn[127] = R0S;

    for (let i = 126; i >= 1; i--) {
      dn = sqrt(-2.0 * log(vn / dn + exp(-0.5 * dn * dn)));
      this.kn[i + 1] = ~~(dn / tn * m1);
      tn = dn;
      this.fn[i] = exp(-0.5 * dn * dn);
      this.wn[i] = dn / m1;
    }
  }

  next() {
    return this._randSample() * this.stdev + this.mean;
  }

  _randSample() {
    let hz = this._xorshift(),
        iz = hz & 127;

    return abs(hz) < this.kn[iz] ? hz * this.wn[iz] : this._nFix(hz, iz);
  }

  _nFix(hz, iz) {
    let r = R0,
        r1 = R1,
        x,
        y;

    while (true) {
      x = hz * this.wn[iz];

      if (iz === 0) {
        do {
          x = -log(this._uni()) * r1;
          y = -log(this._uni());
        } while (y + y < x * x); // {
        //   x = -log(this._uni()) * r1
        //   y = -log(this._uni())
        // }


        return hz > 0 ? r + x : -r - x;
      }

      if (this.fn[iz] + this._uni() * (this.fn[iz - 1] - this.fn[iz]) < exp(-0.5 * x * x)) return x;
      hz = this._xorshift();
      iz = hz & 127;
      if (abs(hz) < this.kn[iz]) return hz * this.wn[iz];
    }
  }

  _xorshift() {
    let m = this.jsr,
        n = m;
    n ^= n << 13;
    n ^= n >>> 17;
    n ^= n << 5;
    this.jsr = n;
    return m + n | 0;
  }

  _uni() {
    return 0.5 + this._xorshift() / N2P32;
  }

}

class Roulett {
  /**
   * From [min, max] return a random integer.
   * Of [min, max], both min and max are inclusive.
   * @param {number} min(inclusive) - int
   * @param {number} max(inclusive) - int
   * @returns {number} int
   */
  static between(min, max) {
    return ~~(random() * (max - min + 1)) + min;
  }
  /**
   * From [min, max) return a random integer.
   * Of [min, max), min is inclusive but max is exclusive.
   * @param {number} min(inclusive) - int
   * @param {number} max(exclusive) - int
   * @returns {number} int
   */


  static rand(min, max) {
    return ~~(random() * (max - min)) + min;
  }

  static index(len) {
    return ~~(random() * len);
  }

  static arrayIndex(arr) {
    return randIdx(arr);
  }

  static element(arr) {
    return randSel(arr);
  }

  static key(obj) {
    return randSel(Object.keys(obj));
  }

  static value(obj) {
    return randSel(Object.values(obj));
  }

  static entry(obj) {
    return randSel(Object.entries(obj));
  }
  /**
   * Return an array of random number that follows gaussian distribution(normal distribution).
   * @param {number} len - length of the returned array
   * @param {number} mean - mean
   * @param {number} stdev - standard deviation
   * @returns {number[]}
   */


  static gaussSamples(len, mean = 0, stdev = 1) {
    const z = new Ziggurat(mean, stdev),
          arr = Array(len);

    for (let i = 0; i < len; i++) arr[i] = z.next();

    return arr;
  }

}

const expReg = /[A-Z]+|[0-9]+/g;

const sepWords = jvExp => jvExp.replace(expReg, it => ' ' + it.toLowerCase()).trim();
/**
 *
 * @param {string} name
 * @returns {string}
 */


const readify = name => {
  var _name;

  return (_name = name, sepWords(_name)).replace('light ', 'l.').replace('deep ', 'd.');
};

const toDataPicker = space => {
  switch (space) {
    case 'rgb':
      return x => {
        var _x;

        return _x = x, farbe.Rgb.fromHex(_x);
      };

    case 'hsl':
      return x => {
        var _ref, _x2;

        return _ref = (_x2 = x, farbe.Rgb.fromHex(_x2)), farbe.Rgb.toHsl(_ref);
      };

    case 'hex':
    default:
      return x => x;
  }
};

function toStatistic(space, cellColor = false) {
  switch (space) {
    case 'rgb':
    case 'hsl':
      return arr => {
        let [a, b, c] = [0, 0, 0],
            len = arr.length;
        veho.Ar.map(arr, ([x, y, z]) => [a, b, c] = [a + x, b + y, c + z], len);
        return [borel.Zu.round(a / len), borel.Zu.round(b / len), borel.Zu.round(c / len)];
      };

    case 'hex':
      return arr => {
        var _ref2, _arr$map;

        return _ref2 = (_arr$map = arr.map(farbe.Rgb.fromHex), toStatistic('rgb')(_arr$map)), farbe.Rgb.toHex(_ref2);
      };

    default:
      return x => x;
  }
}

const toFormatter = (space, colorify = false) => {
  var _space;

  let formatter = space === 'rgb' || space === 'hsl' ? vec => vec.map(x => String(borel.Zu.round(x)).padStart(3)) : null;
  let hatsu = colorify ? (_space = space, toHatsu(_space)) : null;
  return formatter ? hatsu ? x => {
    var _ref3, _x3, _x4;

    return _ref3 = (_x3 = x, formatter(_x3)), (_x4 = x, hatsu(_x4))(_ref3);
  } : formatter : hatsu ? x => {
    var _x5, _x6;

    return _x5 = x, (_x6 = x, hatsu(_x6))(_x5);
  } : x => x;
};

function toHatsu(space) {
  switch (space) {
    case 'rgb':
      return rgb => hatsu.Hatsu.rgb(rgb).inverse;

    case 'hsl':
      return hsl => hatsu.Hatsu.hsl(hsl).inverse;

    case 'hex':
      return hex => hatsu.Hatsu.hex(hex).inverse;

    default:
      return null;
  }
}

const accents = veho.Ar.ini(4, i => `accent_${i + 1}`).reverse(),
      lightens = veho.Ar.ini(5, i => `lighten_${i + 1}`).reverse(),
      darkens = veho.Ar.ini(4, i => `darken_${i + 1}`);
const Degrees = {
  entire: [...accents, ...lightens, 'base', ...darkens],
  base: ['base'],
  lightens: lightens,
  darkens: darkens,
  accents: accents,
  readable: [...accents.slice(-3), ...lightens.slice(-3), 'base']
};
const red = ['red', 'pink'];
const purple = ['purple', 'deepPurple'];
const blue = ['indigo', 'blue', 'lightBlue', 'cyan'];
const green = ['teal', 'green'];
const yellowGreen = ['lightGreen', 'lime', 'yellow'];
const orange = ['amber', 'orange', 'deepOrange'];
const grey = ['brown', 'blueGrey', 'grey'];
const rainbow = [].concat(red, purple, blue, green, yellowGreen, orange);
const entire = rainbow.concat(grey);
const ColorGroups = {
  red,
  purple,
  blue,
  green,
  yellowGreen,
  orange,
  grey,
  rainbow,
  entire
};
const oc = Object.prototype.toString;
/**
 *
 * @param {*} o
 * @return {*}
 */

function clone(o) {
  if (!o || typeof o !== 'object') return o;

  switch (oc.call(o).slice(8, 11)) {
    case 'Arr':
      return dpArr(o);

    case 'Obj':
      return dpObj(o);

    case 'Map':
      return dpMap(o);

    case 'Dat':
      return new Date(+o);

    case 'Set':
      return new Set(dpArr([...o]));
  }

  throw new Error('Unable to copy obj. Unsupported type.');
}
/**
 *
 * @param {Map<*, *>} o
 * @return {Map<*, *>}
 */


function dpMap(o) {
  return new Map([...o.entries()].map(([k, v]) => [k, clone(v)]));
}
/**
 *
 * @param {*[]} o
 * @return {*[]}
 */


function dpArr(o) {
  return o.map(clone);
}
/**
 * Known issue:
 * Unable to clone circular and nested object.
 * @param {{}} o
 * @return {{}}
 */


function dpObj(o) {
  const x = {};

  for (let [k, v] of Object.entries(o)) x[k] = clone(v);

  return x;
} // let protoType = function (it) {
//   const raw = typeof it;
//   if (raw === "object") {
//     switch (true) {
//       case (it instanceof Array):
//         return "array";
//       case (it instanceof Map):
//         return "map";
//       case (it instanceof Set):
//         return "set";
//       case (it instanceof Function):
//         return "function";
//       default:
//         return raw;
//     }
//   } else {
//     return raw
//   }
// };


const oc$1 = Object.prototype.toString;
/**
 * const rxObj = /^\[object (.*)]$/
 * Equivalent to: Object.prototype.stringify.call(o).match(rxObj)[1]
 * @param {*} o
 * @return {string}
 */

const otype = o => oc$1.call(o).slice(8, -1);

const NUM = 'number';
const STR = 'string';
const OBJ = 'object';
/**
 * validate
 * @param x
 * @param y
 * @returns {number}
 */

const vdt = (x, y) => isNaN(x - y) ? NaN : y;

class Num {
  // Angular 4.3
  static isNumeric(x) {
    return !isNaN(x - parseFloat(x));
  }

  static numeric(x) {
    return vdt(x, parseFloat(x));
  }

  static inferData(x) {
    const t = typeof x;
    return t === STR ? isNaN(x - parseFloat(x)) ? STR : NUM : t === OBJ ? otype(x).toLowerCase() : t;
  }

}

const check = x => !!x || x === 0;

class NumLoose {
  static isNumeric(x) {
    return check(+x);
  }

  static numeric(x) {
    x = +x;
    return check(x) ? x : NaN;
  }
  /**
   *
   * @param {*} x
   * @return {string}
   */


  static inferData(x) {
    const t = typeof x;
    return t === STR ? check(+x) ? NUM : STR : t === OBJ ? otype(x).toLowerCase() : t;
  }

}
/**
 * Static class containing methods create 1d-array.
 */


const {
  numeric: num
} = Num,
      {
  numeric: numLoose
} = NumLoose;

class Ar {
  /**
   * Create an array.
   * @param {number} size Integer starts at zero.
   * @param {function(number):*|*} [ject] Defines the how index i decides value(i).
   * @returns {number[]} The
   */
  static ini(size, ject) {
    if (size <= 128) {
      let arr = [];

      if (typeof ject === 'function') {
        for (let i = 0; i < size; i++) arr[i] = ject(i);
      } else {
        for (let i = 0; i < size; i++) arr[i] = ject;
      }

      return arr;
    } else {
      return typeof ject === 'function' ? Array(size).fill(null).map((_, i) => ject(i)) : Array(size).fill(ject);
    }
  }

  static isEmpty(arr) {
    return !arr || !arr.length;
  }
  /**
   *
   * @param {Array} arr
   * @param {boolean} [loose]=false
   * @returns {*}
   */


  static numeric(arr, {
    loose = false
  }) {
    return arr.map(loose ? numLoose : num);
  }

  static clone(arr) {
    return dpArr(arr);
  }

  static indexes(arr) {
    return arr.map((_, i) => i);
  }

  static map(arr, fn, l) {
    l = l || arr.length;
    const vc = Array(l);

    for (--l; l >= 0b0; l--) vc[l] = fn(arr[l], l);

    return vc;
  }

  static mutateMap(arr, fn, l) {
    l = l || arr.length;

    for (--l; l >= 0b0; l--) arr[l] = fn(arr[l], l);

    return arr;
  }

  static select(arr, indexes, hi) {
    hi = hi || indexes.length;
    const vc = Array(hi);

    for (--hi; hi >= 0b0; hi--) vc[hi] = arr[indexes[hi]];

    return vc;
  }
  /**
   *
   * @param {*[]} arr
   * @param {number[]} indexes - number indexes of the positions to be spliced, should be in ascending order.
   * @param {number} [hi]
   * @returns {*[]}
   */


  static splices(arr, indexes, hi) {
    hi = hi || indexes.length;

    for (--hi; hi >= 0b0; hi--) arr.splice(indexes[hi], 1);

    return arr;
  }
  /**
   * Returns an array built from the elements of a given set of arrays.
   * Each element of the returned array is determined by elements from every one of the array-set with the same index.
   * The returned array has length of the first array in the array set.
   * @param {function} zipper The function {x(i)|x(i) ∈ array(i), i ∈ 0,1...n-1, n=arraySet.size} -> y(i), where y
   * is the returned array
   * @param {*[][]} arraySet The array-set to determine the returned array.
   * @returns {*[]|undefined} array
   */


  static multiZip(zipper, ...arraySet) {
    const firstArray = arraySet[0];

    if (!!firstArray) {
      const [len, cnt] = [firstArray.length, arraySet.length];
      const result = Array(len);

      for (let i = 0; i < len; i++) {
        const params = Array(cnt);

        for (let j = 0; j < cnt; j++) {
          params[j] = arraySet[j][i];
        }

        result[i] = zipper(params);
      }

      return result;
    } else {
      return undefined;
    }
  }
  /**
   *
   * @param {number} size
   * @param {*} initial
   * @param {function} progress
   * @returns {*[]}
   */


  static progression(size, initial, progress) {
    switch (size) {
      case 0:
        return [];

      case 1:
        return [initial];

      default:
        const arr = new Array(size);
        arr[0] = initial;

        for (let i = 1; i < size; i++) {
          arr[i] = progress(arr[i - 1]);
        }

        return arr;
    }
  }

  static range(min, max, fn) {
    const len = max - min + 1,
          arr = new Array(len);

    if (fn && typeof fn === 'function') {
      for (let i = 0; i < len; i++) arr[i] = fn(min + i);
    } else {
      for (let i = 0; i < len; i++) arr[i] = min + i;
    }

    return arr;
  }
  /**
   * Create an arithmetic progression
   * @param {number} size
   * @param {number|string} initial
   * @param {number} delta
   * @returns {*[]}
   */


  static arithmetic(size, initial, delta) {
    return Ar.progression(size, initial, previous => previous + delta);
  }
  /**
   * Create a geometric progression
   * @param {number} size
   * @param {number} initial
   * @param {number} ratio
   * @returns {*[]}}
   */


  static geometric(size, initial, ratio) {
    return Ar.progression(size, initial, previous => previous * ratio);
  }
  /**
   *
   * @param {*[]} ar1
   * @param {*[]} ar2
   * @param {function} product
   * @returns {*[]}
   */


  static decartes(ar1, ar2, product) {
    const l1 = ar1.length,
          l2 = ar2.length;
    let arr = Array(l1 * l2);

    for (let i = 0, j, k = 0; i < l1; i++) {
      for (j = 0; j < l2; j++) {
        arr[k++] = product(ar1[i], ar2[j]);
      }
    }

    return arr; // for (let x of ar1) {
    //   arr.push(...ar2.map(y => product(x, y)))
    // }
  }

  static randSample(arr) {
    // const len = arr.length
    // switch (len) {
    //   case 0:
    //     return undefined
    //   case 1:
    //     return arr[0]
    //   default:
    //     const idx = Roulett.rand(0, len)
    //     return arr[idx]
    // }
    return arr[~~(Math.random() * arr.length)];
  }

  static take(arr, len) {
    return arr.slice(0, len);
  }

  static zip(arL, arR, zipper, l) {
    l = l || arL.length;
    const vc = Array(l);

    for (--l; l >= 0; l--) vc[l] = zipper(arL[l], arR[l], l);

    return vc;
  }

} // Array.prototype.zip = function (another, zipper) {


class PalettTable {
  static meta() {
    return {
      colors: ColorGroups.entire,
      degrees: Degrees.entire,
      colorGroups: Object.keys(ColorGroups),
      degreeGroups: Object.keys(Degrees)
    };
  }
  /**
   *
   * @param {string} [space='hex]
   * @param {string[]} [degrees=Degrees.entire]
   * @param {string[]} [colors=ColorGroups.entire]
   * @param {boolean} [average=false]
   * @param {boolean} [cellColor=false]
   * @returns {CrosTab|{side:string,banner:string,matrix:*[][]}}
   */


  static crosTab({
    space = 'hex',
    degrees = Degrees.entire,
    colors = ColorGroups.entire,
    average = false,
    cellColor = false
  } = {}) {
    var _space, _space2, _Ob$selectValues$map;

    const formatter = (_space = space, toFormatter(_space, cellColor));
    const dataPicker = (_space2 = space, toDataPicker(_space2));
    const matrix = (_Ob$selectValues$map = veho.Ob.selectValues(palett.Palett, colors).map(tube => veho.Ob.selectValues(tube, degrees).map(dataPicker)), veho.Mx.transpose(_Ob$selectValues$map));
    const crosTab = new CrosTab(degrees, colors.map(readify), veho.Mx.map(matrix, formatter));

    if (average) {
      var _space3;

      const statistic = (_space3 = space, toStatistic(_space3));

      const formattedStatistic = ar => {
        var _ref, _ar;

        return _ref = (_ar = ar, statistic(_ar)), formatter(_ref);
      };

      const averageOnColumns = veho.Mx.columns(matrix, formattedStatistic);
      crosTab.unshiftRow('average', averageOnColumns);
      const averageOnRows = [''].concat(Ar.map(matrix, formattedStatistic));
      crosTab.unshiftCol('average', averageOnRows);
    }

    return crosTab;
  }

}

function _defineProperty$3(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const xPalett = PalettTable.crosTab({
  degrees: Degrees.readable,
  colors: ColorGroups.rainbow
});

class PalettSelector {
  static random() {
    const color = Roulett.element(xPalett.banner);
    const degree = Roulett.element(xPalett.side);
    const hex = xPalett.queryCell(degree, color);
    return {
      color,
      degree,
      hex
    };
  }

}

_defineProperty$3(PalettSelector, "pool", xPalett.ht * xPalett.wd);

const Directs = {
  p_wise: 0,
  r_wise: 1,
  c_wise: 2
};

Object.defineProperty(exports, 'Hsl', {
  enumerable: true,
  get: function () {
    return farbe.Hsl;
  }
});
Object.defineProperty(exports, 'Rgb', {
  enumerable: true,
  get: function () {
    return farbe.Rgb;
  }
});
Object.defineProperty(exports, 'greys', {
  enumerable: true,
  get: function () {
    return palett.Greys;
  }
});
Object.defineProperty(exports, 'palette', {
  enumerable: true,
  get: function () {
    return palett.Palett;
  }
});
Object.defineProperty(exports, 'shades', {
  enumerable: true,
  get: function () {
    return palett.Shades;
  }
});
exports.ColorCards = PalettTable;
exports.Directs = Directs;
exports.Pal = Pal;
exports.Visual = Visual;
